#import "TABookmarkController.h"

#define BookmarksKey @"Bookmarks"

@implementation TABookmarkController

NSMutableArray *bookmarks;

+ (NSMenu *) bookmarkMenu
{
	static NSMenu *bookmarkMenu = nil;

	if(!bookmarkMenu)
	{
		bookmarkMenu = [[NSMenu allocWithZone: [NSMenu zone]]
			initWithTitle: @"Bookmarks"];
	}
	return bookmarkMenu;
}

+ (void) addBookmarkMenuTo: (NSMenu *) parentMenu
{
	NSEnumerator *enumerator;
	NSString *bookmark;
	NSMenuItem *item;
	
	item = [parentMenu addItemWithTitle: @"Bookmarks"
		action:NULL keyEquivalent: @""];
	[parentMenu setSubmenu: [self bookmarkMenu] forItem: item];

	bookmarks = [[NSMutableArray alloc] initWithArray:
		[self readBookmarksFromDefaults]];

	item = [[self bookmarkMenu] addItemWithTitle: @"Add Bookmark"
		action: @selector(addBookmark:) keyEquivalent: @"B"];
	[item setTarget: [TABookmarkController class]];

	item = [[self bookmarkMenu] addItemWithTitle: @"Remove Bookmark"
		action: @selector(removeBookmark:) keyEquivalent: @"W"];
	[item setTarget: [TABookmarkController class]];

	[[self bookmarkMenu] addItem: [NSMenuItem separatorItem]];
	
	enumerator = [bookmarks objectEnumerator];
	while((bookmark = [enumerator nextObject]))
	{
		item = [[self bookmarkMenu]
			addItemWithTitle: [bookmark lastPathComponent]
			action: @selector(openBookmark:) keyEquivalent: @""];
		[item setRepresentedObject: bookmark];
		[item setTarget: [TABookmarkController class]];
	}
}

+ (NSArray *) readBookmarksFromDefaults
{
	NSUserDefaults *defaults;
	NSArray *val;
	
	defaults = [NSUserDefaults standardUserDefaults];
	
	val = [defaults objectForKey: BookmarksKey];
	
	return val;
}

+ (void) writeBookmarksToDefaults
{
	[[NSUserDefaults standardUserDefaults] 
		setObject: bookmarks
		forKey: BookmarksKey];
}

+ (void) addBookmark: (id)sender
{
	NSString *path;
	BOOL isDirectory;
	
	isDirectory = NO;
	path = [[[NSApplication sharedApplication] mainWindow] representedFilename];
		
	if(path && [[NSFileManager defaultManager] fileExistsAtPath: path
		isDirectory: &isDirectory] && !isDirectory)
	{
		NSMenuItem *item;
		
		item = [[self bookmarkMenu] addItemWithTitle: [path lastPathComponent]
			action: @selector(openBookmark:) keyEquivalent: @""];
		[item setRepresentedObject: path];
		[item setTarget: [TABookmarkController class]];
		[bookmarks addObject: path];
	}
	[self writeBookmarksToDefaults];
}

+ (void) removeBookmark: (id)sender
{
	NSString *path;

	path = [[[NSApplication sharedApplication] mainWindow] representedFilename];
		
	if(path)
	{
		NSMenuItem *item;
		
		item = [[self bookmarkMenu] itemWithTitle: [path lastPathComponent]];
		if(item)
		{
			[[self bookmarkMenu] removeItem: item];
		}
		[bookmarks removeObject: path];
	}
	[self writeBookmarksToDefaults];
}

+ (void) openBookmark: (id)sender
{
	NSString *bookmarkPath;
	BOOL isDirectory;
	id appDelegate;
		
	isDirectory = NO;
	appDelegate = [NSApp delegate];
	bookmarkPath = [sender representedObject];
	
	if ([[NSFileManager defaultManager] fileExistsAtPath: bookmarkPath
		isDirectory: &isDirectory])
	{
		if (appDelegate && [appDelegate
			respondsToSelector:@selector(application:openFile:)])
		{
			[appDelegate application: NSApp openFile: bookmarkPath];
		}
		else
		{
			[[NSDocumentController sharedDocumentController]
				openDocumentWithContentsOfFile: bookmarkPath display:YES];
		}
	}
}

@end
