#import "TAPreferencesController.h"

#define MatchHTMLTags @"MatchHTMLTags"
#define SelectOutward @"SelectOutward"
#define ParseHTML @"ParseHTML"

NSString *TAPreferencesDidChangeNotification = @"TAPreferencesDidChange";


@implementation TAPreferencesController

+ (id) sharedController
{
	static TAPreferencesController *sharedInstance = nil;
	
	if(!sharedInstance)
	{
		sharedInstance = [[TAPreferencesController alloc] init];
	}
	
	return sharedInstance;
}

- (id)init
{
	if ((self = [super initWithWindowNibName:@"Preferences" owner:self]))
	{
		[self setWindowFrameAutosaveName:@"TextAdditionsPreferences"];
		[self readDefaults];
	}
	
	return self;
}

+ (BOOL) booleanDefaultNamed: (NSString *) name fromDefaults: (NSUserDefaults *) defaults
{
	id value;
	
	value = [defaults objectForKey: name];
	if([value isKindOfClass: [NSString class]])
	{
		return [[(NSString *)value lowercaseString] hasPrefix: @"y"] ? YES : NO;
	}
	else if([value isKindOfClass: [NSNumber class]])
	{
		return [(NSNumber *)value boolValue];
	}
	return NO;
}

- (void)readDefaults
{
	NSUserDefaults *defaults;

	defaults = [NSUserDefaults standardUserDefaults];

	matchHTMLTags = [[self class] booleanDefaultNamed: MatchHTMLTags fromDefaults: defaults];
	selectOutward = [[self class] booleanDefaultNamed: SelectOutward fromDefaults: defaults];
	parseHTML = [[self class] booleanDefaultNamed: ParseHTML fromDefaults: defaults];
}


- (BOOL) matchHTMLTags
{
	return matchHTMLTags;
}

- (BOOL) selectOutward
{
	return selectOutward;
}

- (BOOL) parseHTML
{
	return parseHTML;
}

- (void)didChange
{
	[[NSNotificationCenter defaultCenter]
		postNotificationName: TAPreferencesDidChangeNotification object: self];
}

- (void)revertPanel
{    
	[[checkBoxMatrix cellAtRow: MatchHTMLTagsCheckBoxRow column: 0]
		setState: [self matchHTMLTags]];
	[[checkBoxMatrix cellAtRow: SelectOutwardCheckBoxRow column: 0]
		setState: [self selectOutward]];
	[[checkBoxMatrix cellAtRow: ParseHTMLCheckBoxRow column: 0]
		setState: [self parseHTML]];
}


- (IBAction)checkBoxMatrixAction: (id)sender
{
	// Called when a checkbox changes state.
	switch ([sender selectedRow])
	{
		case MatchHTMLTagsCheckBoxRow:
			matchHTMLTags = [[sender selectedCell] state];
			[[NSUserDefaults standardUserDefaults] 
				setObject: (matchHTMLTags ? @"YES" : @"NO") 
				forKey: MatchHTMLTags];
			break;
		case SelectOutwardCheckBoxRow:
			selectOutward = [[sender selectedCell] state];
			[[NSUserDefaults standardUserDefaults] 
				setObject: (selectOutward ? @"YES" : @"NO")
				forKey: SelectOutward];
			break;
		case ParseHTMLCheckBoxRow:
			parseHTML = [[sender selectedCell] state];
			[[NSUserDefaults standardUserDefaults] 
				setObject: (parseHTML ? @"YES" : @"NO")
				forKey: ParseHTML];
			break;
	}
	[self didChange];
}

- (void)windowDidLoad
{
	[super windowDidLoad];
	[self revertPanel];
}

@end
