/* TextDelegate.m created by rklems on Fri 15-Oct-1999 */

#import "TATextDelegate.h"
#import "TATextUtils.h"

#import "TAAboutPanelController.h"
#import "TAPreferencesController.h"
#import "TABookmarkController.h"
#import "TA_NSDictionaryExtensions.h"


@implementation TATextDelegate

static BOOL TA_observingTextViewDidChangeSelection = NO;

+ (void) updateObserversFromPreferences: (NSNotification *) notification
{
	TAPreferencesController *preferences;
	
	preferences = [TAPreferencesController sharedController];
	if([preferences matchHTMLTags] || [preferences selectOutward]
		|| [preferences parseHTML])
	{
		if(!TA_observingTextViewDidChangeSelection)
		{
			[[NSNotificationCenter defaultCenter] addObserver: self
				selector: @selector(textViewSelectionDidChange:)
				name: NSTextViewDidChangeSelectionNotification
				object: nil];
			TA_observingTextViewDidChangeSelection = YES;
		}
	}
	else
	{
		if(TA_observingTextViewDidChangeSelection)
		{
			[[NSNotificationCenter defaultCenter] removeObserver: self
				name: NSTextViewDidChangeSelectionNotification
				object: nil];
			TA_observingTextViewDidChangeSelection = NO;
		}
	}
}

+ (void) load
{
	[self updateObserversFromPreferences:nil];
	[[NSNotificationCenter defaultCenter] addObserver:self
		selector:@selector(updateObserversFromPreferences:)
		name: TAPreferencesDidChangeNotification
		object: nil];
	[self performSelector: @selector(insertMenu:)
		withObject:nil afterDelay:0.0];
}

+ (void) stepOutward:(id)timer
{
	NSTextView *textView;
	NSRange braceRange;
	NSRange theRange;

	if([[NSApp currentEvent] modifierFlags] & NSAlternateKeyMask)
	{
		BOOL newTimer = NO;
		
		if([timer isKindOfClass: [NSTimer class]])
		{
			textView = (NSTextView *)[timer userInfo];
		}
		else
		{
			textView = (NSTextView *)[timer object];
		}
		braceRange = theRange = [textView selectedRange];
		
		// Step out one level of braces...
		braceRange = TA_findBraceNextBrace(theRange, [textView string]);
	
		if(braceRange.location != NSNotFound)
		{
			theRange = braceRange;
			newTimer = YES;
		}
		else
		{
			theRange = NSMakeRange(0, ([[textView string] length] - 1));
		}
		
		[textView setSelectedRange: theRange 
			affinity: NSSelectionAffinityUpstream stillSelecting: YES];
		[textView scrollRangeToVisible: theRange];
		if(newTimer)
		{
			[NSTimer scheduledTimerWithTimeInterval: .5
				target: self
				selector: @selector(stepOutward:)
				userInfo: textView
				repeats: NO];
		}
	}
}

+ (void) handleLeftMouseUp: (NSNotification *)notification
{
	TAPreferencesController *preferences;
	NSEvent *event;
	
	preferences = [TAPreferencesController sharedController];
	event = [NSApp currentEvent];
	
	if([preferences selectOutward] && 
		[event modifierFlags] & NSAlternateKeyMask)
	{
		[self stepOutward: notification];
	}
	else if([preferences matchHTMLTags] && [event clickCount] == 2)
	{
		NSRange matchRange;
		NSRange selRange;
		NSString *lowercaseString;
		NSAutoreleasePool *pool;
		NSTextView *textView;
		Tag tag;
		
		pool = [[NSAutoreleasePool alloc] init];
		
		textView = [notification object];
		selRange = [textView selectedRange];
		lowercaseString = [[textView string] lowercaseString];
		
		tag = TA_findFullRangeOfTagForRangeInString(
			selRange, lowercaseString);
		if(tag.type != TANotTag)
		{
			if(([event modifierFlags] & NSCommandKeyMask) &&
				[preferences parseHTML])
			{
				BOOL attempt = NO;
				
				if(TA_isSingleTag(tag) && TA_isWOAbleTag(tag))
				{
					attempt = YES;
				}
				else
				{
					matchRange = TA_findMatchingTagForTagInString(
						tag, lowercaseString);
					if(matchRange.location != NSNotFound)
					{
						if(tag.type == TACloseTag)
						{
							tag.matchingTagRange = tag.tagRange;
							tag.tagRange = matchRange;
						}
						else
						{
							tag.matchingTagRange = matchRange;
						}
						attempt = YES;
					}
				}
				if(attempt)
				{
					NSString *stringToParse;
					
					if(tag.matchingTagRange.location != NSNotFound)
					{
						stringToParse = [[textView string] substringWithRange: 
							NSUnionRange(tag.tagRange, tag.matchingTagRange)];
					}
					else
					{
						stringToParse = [[textView string] substringWithRange: 
							tag.tagRange];
					}
					TA_parseHTMLStringToWebObject(stringToParse);
				}
			}
			else
			{
				if(TA_isSingleTag(tag)) {
					[textView setSelectedRange: tag.tagRange
						affinity: NSSelectionAffinityUpstream
						stillSelecting: YES];
				}
				else
				{
					matchRange = TA_findMatchingTagForTagInString(
						tag, lowercaseString);
		
					if(matchRange.location != NSNotFound) {
						selRange = NSUnionRange(tag.tagRange, matchRange);
						[textView setSelectedRange: selRange
							affinity: NSSelectionAffinityUpstream
							stillSelecting: YES];
						[textView scrollRangeToVisible:matchRange];
					}
				}
			}
		}
		[pool release];
	}
}

+ (void) textViewSelectionDidChange: (NSNotification *)notification
{
	switch([[NSApp currentEvent] type])
	{
		case NSLeftMouseUp:
			[self handleLeftMouseUp: notification];
			break;
		default:
			break;
	}
}

+ (void)insertMenu: (NSNotification *) notification
{
	NSMenu *mainMenu;

	if((mainMenu = [NSApp mainMenu]))
	{
		NSMenu *insertIntoMenu;
		NSMenuItem *item;
		
		if((item = [mainMenu itemWithTitle: @"Format"]) &&
			[item hasSubmenu])
		{
			unsigned int insertLocation;
			
			insertIntoMenu = [item target];
			item = [insertIntoMenu itemWithTitle: @"TextExtras"];
			
			if(!item)
			{
				// No TextExtras, go below the "Text" menu
				item = [insertIntoMenu itemWithTitle: @"Text"];
			}
			
			if(item)
			{
				insertLocation = [[insertIntoMenu itemArray]
					indexOfObjectIdenticalTo: item] + 1;
			}
			else
			{
				// Otherwise, just stick it at the bottom of the pile
				insertLocation = [[insertIntoMenu itemArray] count];
			}
			
			if(insertIntoMenu)
			{
				NSMenu *extrasMenu;
				
				extrasMenu = [[NSMenu allocWithZone: [NSMenu zone]]
					initWithTitle: @"TextAdditions"];
				item = [insertIntoMenu insertItemWithTitle: @"TextAdditions"
					action: NULL keyEquivalent: @"" atIndex: insertLocation];
				[insertIntoMenu setSubmenu: extrasMenu forItem: item];

				[TABookmarkController addBookmarkMenuTo: extrasMenu];

				item = [extrasMenu addItemWithTitle: @"About TextAdditions"
					action:@selector(showWindow:) keyEquivalent: @""];
				[item setTarget: [TAAboutPanelController sharedController]];
				
				item = [extrasMenu addItemWithTitle: @"Preferences"
					action:@selector(showWindow:) keyEquivalent: @""];
				[item setTarget: [TAPreferencesController sharedController]];
				
				item = [extrasMenu addItemWithTitle: @"Parse HTML to WO"
					action:@selector(TA_parseHTMLToWebObjects:)
					keyEquivalent: @""];
			}
		}
	}
}

@end
