#import "TATextUtils.h"
#import "TA_NSDictionaryExtensions.h"

@interface NSTextView(TAHTMLToWebObjects)
- (void) TA_parseHTMLToWebObjects:(id) sender;
@end

@implementation NSTextView(TAHTMLToWebObjects)

- (void) TA_parseHTMLToWebObjects:(id) sender
{
	TA_parseHTMLStringToWebObject([self string]);
}

@end

void TA_parseHTMLStringToWebObject(NSString *string)
{
	Class documentClass;
	
	documentClass = NSClassFromString(@"Document");
	if(documentClass && 
		[documentClass respondsToSelector:@selector(openUntitled)])
	{
		id wod, wos, html;
		NSTextView *wodView, *wosView, *htmlView;
		NSMutableString *wosString, *wodString;
		NSAutoreleasePool *pool;
		
		html = objc_msgSend(documentClass, @selector(openUntitled));
		htmlView = [html firstTextView];
		[htmlView insertText: string];

		wodString = [[NSMutableString alloc] init];
		wosString = [[NSMutableString alloc] init];
		
		pool = [[NSAutoreleasePool alloc] init];
		
		while(1)
		{
			NSAutoreleasePool *innerPool;
			NSString *lowercaseString;
			Tag tag;
			
			innerPool = [[NSAutoreleasePool alloc] init];
			
			lowercaseString = [[htmlView string] lowercaseString];
			
			tag = TA_findNextWOAbleTagInStringFromRange(NSMakeRange(0, 0),
				lowercaseString);

			if(tag.type == TANotTag)
			{
				break;
			}

			if(!TA_isSingleTag(tag))
			{
				tag.matchingTagRange = TA_findMatchingTagForTagInString(
					tag, lowercaseString);
			}
			
			TA_ParseTagContents(&tag, [htmlView string]);
			TA_ConvertTagInfoToWebObjectsMapping(&tag);
			[wodString appendString: [tag.tagInfo descriptionAsWod]];
			[wosString insertString: [tag.tagInfo variableDeclarations]
				atIndex: 0];
			
			if([[tag.tagInfo objectForKey:@"WOTagName"]
				isEqualToString:@"WOSubmitButton"])
			{
				[wosString appendString: [tag.tagInfo actionDeclaration]];
			}

			if(tag.deleteBetween ||
				tag.matchingTagRange.location == NSNotFound)
			{
				NSRange fullRange;
				NSString *newString;
				
				if(tag.matchingTagRange.location != NSNotFound)
				{
					fullRange = NSUnionRange(tag.tagRange,
						tag.matchingTagRange);
				}
				else
				{
					fullRange = tag.tagRange;
				}
				newString = [[NSString alloc] initWithFormat:
					@"<WEBOBJECT NAME=\"%@\"></WEBOBJECT>",
					[tag.tagInfo objectForKey:@"WOBindingName"]];
				if([htmlView shouldChangeTextInRange: fullRange
					replacementString: newString])
				{
					[[htmlView textStorage] replaceCharactersInRange: fullRange
						withString: newString];
					[htmlView didChangeText];
				}
				[newString release];
			}
			else
			{
				NSString *newString;
				
				tag.matchingTagRange.location += ((19 + 
					[(NSString *)[tag.tagInfo objectForKey:@"WOBindingName"]
						length]) - tag.tagRange.length);
				
				newString = [[NSString alloc] initWithFormat:
					@"<WEBOBJECT NAME=\"%@\">", 
						[tag.tagInfo objectForKey:@"WOBindingName"]];
				if([htmlView shouldChangeTextInRange: tag.tagRange
					replacementString: newString])
				{
					[[htmlView textStorage] replaceCharactersInRange:
						tag.tagRange withString: newString];
					[htmlView didChangeText];
					if([htmlView shouldChangeTextInRange: tag.matchingTagRange
						replacementString: @"</WEBOBJECT>"])
					{
						[[htmlView textStorage] replaceCharactersInRange:
							tag.matchingTagRange withString: @"</WEBOBJECT>"];
						[htmlView didChangeText];
					}
				}
				[newString release];
			}
			[innerPool release];
		}
		
		wod = objc_msgSend(documentClass, @selector(openUntitled));
		wodView = [wod firstTextView];
		[wodView insertText: wodString];
		[wodString release];
		
		if([wosString length])
		{
			wos = objc_msgSend(documentClass, @selector(openUntitled));
			wosView = [wos firstTextView];
			[wosView insertText: wosString];
		}
		[wosString release];
		
		[pool release];
	}
}

Tag TA_findNextWOAbleTagInStringFromRange(NSRange startRange, NSString *string)
{
	NSRange foundRange;
	NSRange searchRange;
	Tag tag;
	int found = 0;
	
	tag.name = nil;
	tag.tagInfo = nil;
	tag.tagRange = NSMakeRange(NSNotFound, 0);
	tag.matchingTagRange = NSMakeRange(NSNotFound, 0);
	tag.type = TANotTag;
	tag.deleteBetween = NO;
	
	foundRange = startRange;
	while(!found && foundRange.location != NSNotFound)
	{
		Tag innerTag;
		
		searchRange = NSMakeRange(foundRange.location,
			[string length] - foundRange.location);
		foundRange = [string rangeOfString: @"<" 
			options: 0 range: searchRange];
		if(foundRange.location != NSNotFound)
		{
			foundRange.location++;
			innerTag = TA_findFullRangeOfTagForRangeInString(foundRange,
				string);
			if(innerTag.type == TAOpenTag && TA_isWOAbleTag(innerTag))
			{
				found = 1;
				tag = innerTag;
			}
		}
	}
	return tag;
}

NSRange TA_findBraceNextBrace(NSRange startRange, NSString *string)
{
	NSRange newRange;
	unichar downChar = '\0';
	unichar upChar = '\0';
	int length;
	int offset;
	
	newRange = startRange;
	length = [string length];
	offset = 0;
	while((newRange.location + newRange.length) < length)
	{
		unichar chr;
		
		switch((chr = [string characterAtIndex:
			(newRange.location + newRange.length)]))
		{
			case '{' :
						offset--;
						break;
			case '[' :
						offset--;
						break;
			case '(' :
						offset--;
						break;
			case '}' :
						offset++;
						break;
			case ']' :
						offset++;
						break;
			case ')' :
						offset++;
						break;
			default  :
						break;
		}
		newRange.length++;
		if(offset > 0)
		{
			switch(chr)
			{
				case '}' :
							upChar = '}';
							downChar = '{';
							break;
				case ']' :
							upChar = ']';
							downChar = '[';
							break;
				case ')' :
							upChar = ')';
							downChar = '(';
							break;
				default  :
							break;
			}
			break;
		}
	}
	if(offset < 1)
	{
		return NSMakeRange( NSNotFound, 0 );
	}
	offset = 1;
	if(newRange.location > 0)
	{
		// back up a char...
		newRange.location--;
		newRange.length++;
	}
	while(newRange.location)
	{
		unichar chr;
		
		chr = [string characterAtIndex: newRange.location];
		if(chr == downChar)
		{
			offset--;
		}
		else if(chr == upChar)
		{
			offset++;
		}
		if(!offset)
		{
			break;
		}
		newRange.location--;
		newRange.length++;
	}
	if(offset)
	{
		return NSMakeRange( NSNotFound, 0 );
	}
	return newRange;
}

Tag TA_findFullRangeOfTagForRangeInString(NSRange selRange, NSString *string)
{
	Tag returnTag;
	unichar currentChar;
	int stringLength;
	int found;
	int type;
	int nameFound;
	NSRange nameRange;

	returnTag.name = nil;
	returnTag.tagInfo = nil;
	returnTag.tagRange = NSMakeRange(NSNotFound, 0);
	returnTag.matchingTagRange = NSMakeRange(NSNotFound, 0);
	returnTag.type = TANotTag;
	returnTag.deleteBetween = NO;

	nameRange.location = selRange.location;
	nameRange.length = selRange.length;
	
	// Okay, we've got our string, we need to figure out if this is a real
	// tag, and then figure out if its an open or close...
	if(selRange.location == 0) { 
		return returnTag;
	}
    selRange.location--;
	selRange.length++;
	currentChar = [string characterAtIndex: selRange.location];
	if(currentChar != '<')
	{
		if(currentChar != '/' || !selRange.location)
		{
			return returnTag;
		}
		else
		{
			selRange.location--;
			selRange.length++;
			currentChar = [string characterAtIndex: selRange.location];
			if(currentChar != '<')
			{
				return returnTag;
			}
			type = TACloseTag;
		}
	}
	else
	{
		type = TAOpenTag;
	}

	// Okay we know if its an open or close tag now... but still not if its
	// valid.

	stringLength = [string length];
	found = 0;
	nameFound = 0;
	for(;!found && (selRange.location + selRange.length) < stringLength;)
	{
		selRange.length++;
		currentChar = [string characterAtIndex:
			((selRange.location + selRange.length) - 1)];
		switch(currentChar)\
		{
			case '<' :
						return returnTag;
			case '>' :
						found = 1;
						returnTag.type = type;
						returnTag.tagRange = selRange;
						break;
			case ' ' :
						if(!nameFound) {
							nameFound = 1;
						}
			default  :
						if(!nameFound)
						{
							nameRange.length++;
						}
						break;					
		}
	}
	
	// Set the tag's name
	returnTag.name = [string substringWithRange: nameRange];

	// We now have the tag in its entirety
	return returnTag;
}

NSRange TA_findMatchingTagForTagInString(Tag tag, NSString *string)
{
	BOOL backwards;
	int tagBalance;
	NSRange searchRange;
	NSRange foundRange;
	int stringLength;
	Tag loopTag;
	tagBalance = 1;
	stringLength = [string length];

	if(tag.type == TACloseTag)
	{
		backwards = YES;
		searchRange.location = 0;
		searchRange.length = tag.tagRange.location;
	} else {
		backwards = NO;
		searchRange.location = tag.tagRange.location + tag.tagRange.length;
		searchRange.length = stringLength - searchRange.location;
	}

	while(tagBalance > 0)
	{

		foundRange = [string rangeOfString:tag.name 
			options: (backwards ? NSBackwardsSearch : 0)
			range: searchRange];
			
		if(foundRange.location != NSNotFound)
		{
			loopTag = TA_findFullRangeOfTagForRangeInString(foundRange, string);
			
			if(loopTag.type != TANotTag && 
				![loopTag.name isEqualToString: tag.name])
			{
				loopTag.type = TANotTag;
			}
			
			switch(loopTag.type)
			{
				case TACloseTag :
									backwards ? tagBalance++ : tagBalance--;
									break;
				case TAOpenTag  :
									backwards ? tagBalance-- : tagBalance++;
									break;
				default           :
									if(backwards)
									{
										searchRange.location = 0;
										searchRange.length = 
											foundRange.location;
									}
									else
									{
										searchRange.location =
											foundRange.location +
											foundRange.length;
										searchRange.length = stringLength -
											searchRange.location;
									}
									break;
			}
			if(loopTag.type != TANotTag && tagBalance)
			{
				if(backwards)
				{
					searchRange.location = 0;
					searchRange.length = loopTag.tagRange.location;
				}
				else
				{
					searchRange.location = loopTag.tagRange.location +
						loopTag.tagRange.length;
					searchRange.length = stringLength -	searchRange.location;
				}
			}
		}
		else
		{
			break;
		}
	}
	if(!tagBalance)
	{
		return loopTag.tagRange;
	}
	return NSMakeRange(NSNotFound, 0);
}

BOOL TA_isSingleTag(Tag tag)
{
	static NSSet *singleTags = nil;
	
	if(!singleTags)
	{
		singleTags = [[NSSet alloc] initWithObjects:
			@"br", @"hr", @"img", @"meta", @"link", @"input", @"option",
			@"area", @"base", @"basefont", @"bgsound", @"col", @"embed",
			@"frame", @"overlay", @"param", @"tab", nil];
	}
	return [singleTags containsObject: tag.name];
}

BOOL TA_isWOAbleTag(Tag tag)
{
	static NSSet *WOAbleTags = nil;
	
	if(!WOAbleTags)
	{
		WOAbleTags = [[NSSet alloc] initWithObjects:
			@"img", @"input", @"form", @"a", @"select", @"textarea", nil];
	}
	return [WOAbleTags containsObject: tag.name];
}

/*
 * Possible known bug here, not sure how it will handle a key w/o
 * a value (nowrap on td for instance...)
 */
void TA_ParseTagContents(Tag *tag, NSString *string)
{
	const char *valueStartPtr, *valueEndPtr, *endOfTag;
	const char *lookPtr, *equalPtr, *startPtr;
	NSString *substring;
	NSAutoreleasePool *pool;
	NSMutableDictionary *tagInfo;
	
	pool = [[NSAutoreleasePool alloc] init];
	
	tagInfo = [[NSMutableDictionary alloc] init];
	substring = [string substringWithRange: tag->tagRange];
	startPtr = [substring cString];
	endOfTag = startPtr + strlen(startPtr);
	
	startPtr += [tag->name length] + 1; // Scan past <NAME
	lookPtr = startPtr;
	while(lookPtr < endOfTag)
	{
		startPtr = equalPtr = valueStartPtr = valueEndPtr = NULL;
		while(isspace(*lookPtr) && lookPtr < endOfTag)
		{
			lookPtr++;
		}
		if(lookPtr < endOfTag)
		{
			startPtr = lookPtr;
		}
		while(*lookPtr != '=' && lookPtr < endOfTag)
		{
			lookPtr++;
		}
		if(*lookPtr == '=')
		{
			equalPtr = lookPtr;
			if(lookPtr < endOfTag)
			{
				valueStartPtr = ++lookPtr;
				if(*valueStartPtr == '\"')
				{
					if(lookPtr < endOfTag)
					{
						valueStartPtr = ++lookPtr;
						while(*lookPtr != '\"' && lookPtr < endOfTag)
						{
							lookPtr++;
						}
						if(*lookPtr == '\"')
						{
							valueEndPtr = lookPtr;
						}
					}
				}
				else
				{
					while(!isspace(*lookPtr) && lookPtr < endOfTag)
					{
						lookPtr++;
					}
					if(isspace(*lookPtr) || lookPtr == endOfTag)
					{
						valueEndPtr = lookPtr;
					}
				}
			}
		}
		if(startPtr && equalPtr && valueStartPtr && valueEndPtr)
		{
			NSString *key;
			NSString *value;
			
			key = [[NSString alloc] initWithCString: startPtr
				length: (equalPtr - startPtr)];
			value = [[NSString alloc] initWithCString: valueStartPtr
				length: (valueEndPtr - valueStartPtr)];
			[tagInfo setObject: value forKey: [key lowercaseString]];
			[key release];
			[value release];
			lookPtr++;
		}
	}
	[pool release];
	tag->tagInfo = [tagInfo autorelease];
}

static int staticCount = 0;

void TA_doBindingName(NSMutableDictionary *dict, NSString *defaultName)
{
	NSString *bindingName;
	
	bindingName = [[dict objectForKey:@"name"] uppercaseString];
	if(!bindingName)
	{
		bindingName = [NSString stringWithFormat:
			@"%@_%d", defaultName, staticCount++];
	}
	[dict setObject: bindingName forKey:@"WOBindingName"];
}

void TA_ConvertTagInfoToWebObjectsMapping(Tag *tag)
{
	if([tag->name isEqualToString:@"a"])
	{
		[tag->tagInfo setObject:@"WOHyperlink" forKey:@"WOTagName"];
		TA_doBindingName(tag->tagInfo, @"WOHYPERLINK");
	}
	else if([tag->name isEqualToString:@"img"])
	{
		[tag->tagInfo setObject:@"WOImage" forKey:@"WOTagName"];
		TA_doBindingName(tag->tagInfo, @"WOIMAGE");
	}
	else if([tag->name isEqualToString:@"form"])
	{
		[tag->tagInfo setObject:@"WOForm" forKey:@"WOTagName"];
		[tag->tagInfo removeObjectForKey:@"action"];
		TA_doBindingName(tag->tagInfo, @"WOFORM");
	}
	else if([tag->name isEqualToString:@"input"])
	{
		NSString *type;
		
		type = [[tag->tagInfo objectForKey: @"type"] lowercaseString];
		[tag->tagInfo removeObjectForKey:@"type"];
		if([type isEqualToString: @"submit"])
		{	
			[tag->tagInfo setObject:@"WOSubmitButton" forKey:@"WOTagName"];
			TA_doBindingName(tag->tagInfo, @"WOSUBMITBUTTON");
			[tag->tagInfo setObject:
				[NSString stringWithFormat:@"submit%@",
					[[tag->tagInfo objectForKey:@"WOBindingName"]
						capitalizedString]] forKey:@"action"];
		}
		else if([type isEqualToString: @"reset"])
		{
			[tag->tagInfo setObject:@"WOResetButton" forKey:@"WOTagName"];
			TA_doBindingName(tag->tagInfo, @"WORESETBUTTON");
		}
		else if([type isEqualToString: @"hidden"])
		{
			[tag->tagInfo setObject:@"WOHiddenField" forKey:@"WOTagName"];
			TA_doBindingName(tag->tagInfo, @"WOHIDDENFIELD");
			[tag->tagInfo setObject:
				[[tag->tagInfo objectForKey:@"name"] lowercaseString]
				forKey:@"value"];
		}
		else if([type isEqualToString: @"password"])
		{
			[tag->tagInfo setObject:@"WOPasswordField" forKey:@"WOTagName"];
			TA_doBindingName(tag->tagInfo, @"WOPASSWORDFIELD");
			[tag->tagInfo setObject:
				[[tag->tagInfo objectForKey:@"name"] lowercaseString]
				forKey:@"value"];
		}
		else if([type isEqualToString: @"checkbox"])
		{
			[tag->tagInfo setObject:@"WOCheckBox" forKey:@"WOTagName"];
			TA_doBindingName(tag->tagInfo, @"WOCHECKBOX");
			[tag->tagInfo setObject:
				[[tag->tagInfo objectForKey:@"name"] lowercaseString]
				forKey:@"value"];
		}
		else if([type isEqualToString: @"image"])
		{
			[tag->tagInfo setObject:@"WOImageButton" forKey:@"WOTagName"];
			TA_doBindingName(tag->tagInfo, @"WOIMAGEBUTTON");
			[tag->tagInfo setObject:
				[NSString stringWithFormat:@"submit%@",
					[[tag->tagInfo objectForKey:@"WOBindingName"]
						capitalizedString]] forKey:@"action"];
			[tag->tagInfo setObject:
				[[tag->tagInfo objectForKey:@"name"] lowercaseString]
				forKey:@"value"];
		}
		else if([type isEqualToString: @"radio"])
		{
			[tag->tagInfo setObject:@"WORadioButton" forKey:@"WOTagName"];
			TA_doBindingName(tag->tagInfo, @"WORADIOBUTTON");
			[tag->tagInfo setObject:
				[NSString stringWithFormat:@"selected%@",
					[[tag->tagInfo objectForKey:@"WOBindingName"]
						capitalizedString]] forKey:@"selection"];
			[tag->tagInfo setObject:
				[[tag->tagInfo objectForKey:@"name"] lowercaseString]
				forKey:@"value"];
		}
		else if([type isEqualToString: @"text"])
		{
			[tag->tagInfo setObject:@"WOTextField" forKey:@"WOTagName"];
			TA_doBindingName(tag->tagInfo, @"WOTEXTFIELD");
			[tag->tagInfo setObject:
				[[tag->tagInfo objectForKey:@"name"] lowercaseString]
				forKey:@"value"];
		}
	}
	else if([tag->name isEqualToString:@"select"])
	{
		NSString *commonName;
		
		tag->deleteBetween = YES;
		if([tag->tagInfo objectForKey:@"multiple"])
		{
			[tag->tagInfo setObject:@"WOBrowser" forKey:@"WOTagName"];
			TA_doBindingName(tag->tagInfo, @"WOBROWSER");
			commonName = [[tag->tagInfo objectForKey:@"WOBindingName"]
				lowercaseString];
			[tag->tagInfo setObject:
				[NSString stringWithFormat:@"selected%@s", 
					[commonName capitalizedString]] forKey:@"selections"];
		}
		else
		{
			[tag->tagInfo setObject:@"WOPopUpButton" forKey:@"WOTagName"];
			TA_doBindingName(tag->tagInfo, @"WOPOPUPBUTTON");
			commonName = [[tag->tagInfo objectForKey:@"WOBindingName"]
				lowercaseString];
			[tag->tagInfo setObject:
				[NSString stringWithFormat:@"selected%@", 
					[commonName capitalizedString]] forKey:@"selection"];
		}
		[tag->tagInfo setObject:
			[NSString stringWithFormat:@"%@s", commonName] forKey:@"list"];
		[tag->tagInfo setObject:
			[NSString stringWithFormat:@"%@", commonName] forKey:@"item"];
	}
	else if([tag->name isEqualToString:@"textarea"])
	{
		[tag->tagInfo setObject:@"WOText" forKey:@"WOTagName"];
		TA_doBindingName(tag->tagInfo, @"WOTEXT");
	}
	else
	{
		// You get nothing...
	}
}
